/****************************************************************************************
*  YAROCK                                                                               *
*  Copyright (c) 2010-2013 Sebastien amardeilh <sebastien.amardeilh+yarock@gmail.com>   *
*                                                                                       *
*  This program is free software; you can redistribute it and/or modify it under        *
*  the terms of the GNU General Public License as published by the Free Software        *
*  Foundation; either version 2 of the License, or (at your option) any later           *
*  version.                                                                             *
*                                                                                       *
*  This program is distributed in the hope that it will be useful, but WITHOUT ANY      *
*  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A      *
*  PARTICULAR PURPOSE. See the GNU General Public License for more details.             *
*                                                                                       *
*  You should have received a copy of the GNU General Public License along with         *
*  this program.  If not, see <http://www.gnu.org/licenses/>.                           *
*****************************************************************************************/


#include "AlbumCover/coverprovider_discogs.h"
#include "MediaItem/mediaitem.h"
#include "networkaccess.h"
#include "utilities.h"
#include "constants.h" // discogs api key
#include "debug.h"

#include <QtGui>
#include <QtNetwork>



/*******************************************************************************
   API
    http://www.discogs.com/help/api
    http://www.discogs.com/developers/
*******************************************************************************/


/*
********************************************************************************
*                                                                              *
*    Class CoverProvider_Discogs                                               *
*                                                                              *
********************************************************************************
*/
CoverProvider_Discogs::CoverProvider_Discogs(QObject* parent)
  : CoverProvider("discogs", parent)
{
}

//! ----  set_search -----------------------------------------------------------
void CoverProvider_Discogs::set_search(const QString& artist, const QString& album)
{
    m_album      = album;
    m_artist     = artist;
    m_cover_name = MEDIA::coverName(artist, album);
}

//! ----  start_search ---------------------------------------------------------
void CoverProvider_Discogs::start_search(int id)
{
    //Debug::debug() << " ---- CoverProvider_Discogs -> start_search id" << id;
    if (m_artist.isEmpty())
    {
      emit search_finished(id);
    }
    else
    {
      slot_send_search_request(id);
    }
}


//! ----  slot_send_search_request ---------------------------------------------
void CoverProvider_Discogs::slot_send_search_request(int id)
{
    //Debug::debug() << " ---- CoverProvider_Discogs -> slot_send_search_request";
    QString query = m_artist + "+" + m_album;

    QUrl url("http://api.discogs.com/search");
    url.addQueryItem( "api_key", DISCOGS_API_KEY );
    url.addQueryItem( "page", QString::number( 1 ) );
    url.addQueryItem( "type", "releases" );
    url.addQueryItem( "q", query );
    url.addQueryItem( "f", "xml" );

    QObject *reply = HTTP()->get(url);
    m_list_requests[reply] = id;

    connect(reply, SIGNAL(data(QByteArray)), this, SLOT(slot_parse_search_response(QByteArray)));
    connect(reply, SIGNAL(error(QNetworkReply*)), this, SLOT(slot_cover_search_stop()));
}



//! ----  slot_parse_search_response -------------------------------------------
void CoverProvider_Discogs::slot_parse_search_response(QByteArray bytes)
{
    /*-------------------------------------------------*/
    /* Get id from sender reply                        */
    /* ------------------------------------------------*/
    QObject* reply = qobject_cast<QObject*>(sender());
    if (!reply || !m_list_requests.contains(reply)) {
      emit search_finished(-1);
      return;
    }
    const int id = m_list_requests.take(reply);

    /*-------------------------------------------------*/
    /* Parse response                                  */
    /* ------------------------------------------------*/
    QXmlStreamReader xml(bytes);
    while( !xml.atEnd() && !xml.hasError() )
    {
      xml.readNext();
      if( xml.isStartElement() && xml.name() == "searchresults" )
      {
        while( !xml.atEnd() && !xml.hasError() )
        {
          xml.readNext();
          const QStringRef &n = xml.name();
          if( xml.isEndElement() && n == "searchresults" )
            break;
          if( !xml.isStartElement() )
            continue;

          if( n == "result" && xml.attributes().value( "type" ).toString() == "release")
          {
            while( !xml.atEnd() && !xml.hasError() )
            {
              xml.readNext();

              if( xml.isEndElement() && n == "result" )
                break;
              if( !xml.isStartElement() )
                continue;
              if( xml.name() == "uri" )
              {
                QString release_url = xml.readElementText();
                QString releaseId = release_url.split( '/' ).last();
                Debug::debug() << " ---- CoverProvider_Discogs -> releaseId " << releaseId;

                QUrl url("http://api.discogs.com/release/" + releaseId);
                url.addQueryItem( "api_key", DISCOGS_API_KEY );
                url.addQueryItem( "f", "xml" );
                QObject *reply = HTTP()->get(url);
                m_list_requests[reply] = id;
                connect(reply, SIGNAL(data(QByteArray)), this, SLOT(slot_parse_discogs_release(QByteArray)));
                connect(reply, SIGNAL(error(QNetworkReply*)), this, SLOT(slot_cover_search_stop()));
                return;
              }
              else
                xml.skipCurrentElement();
            }
          }
          else
            xml.skipCurrentElement();
        }
      }
    }

    Debug::debug() << " ---- CoverProvider_Discogs -> no release id found";
    emit search_finished(id);
}

//! --  slot_parse_discogs_release ---------------------------------------------
void CoverProvider_Discogs::slot_parse_discogs_release(QByteArray bytes)
{
    //Debug::debug() << " ---- CoverProvider_Discogs -> slot_parse_discogs_release";
    /*-------------------------------------------------*/
    /* Get id from sender reply                        */
    /* ------------------------------------------------*/
    QObject* reply = qobject_cast<QObject*>(sender());
    if (!reply || !m_list_requests.contains(reply))   return;
    const int id = m_list_requests.take(reply);

    /*-------------------------------------------------*/
    /* Parse response                                  */
    /* ------------------------------------------------*/
    QXmlStreamReader xml(bytes);
    while( !xml.atEnd() && !xml.hasError() )
    {
      xml.readNext();
      if( !xml.isStartElement() || xml.name() != "release" )
        continue;
      while( !xml.atEnd() && !xml.hasError() )
      {
        xml.readNext();
        const QStringRef &n = xml.name();
        if( xml.isEndElement() && n == "release" )
          break;
        if( !xml.isStartElement() )
          continue;

        if( n == "images" )
        {
          while( !xml.atEnd() && !xml.hasError() )
          {
            xml.readNext();
            if( xml.isEndElement() && xml.name() == "images" )
              break;
            if( !xml.isStartElement() )
              continue;
            if( xml.name() == "image" )
            {
                const QXmlStreamAttributes &attr = xml.attributes();
                QUrl url( attr.value( "uri" ).toString() );

                QObject *reply = HTTP()->get(url);
                m_list_requests[reply] = id;
                connect(reply, SIGNAL(data(QByteArray)), this, SLOT(slot_cover_received(QByteArray)));
                connect(reply, SIGNAL(error(QNetworkReply*)), this, SLOT(slot_cover_search_stop()));
                return;
            }
            else
              xml.skipCurrentElement();
          }
        }
        else
          xml.skipCurrentElement();
      }
    }

    Debug::debug() << " ---- CoverProvider_Discogs -> no image found";
    emit search_finished(id);
}




//! --  slot_cover_received ----------------------------------------------------
void CoverProvider_Discogs::slot_cover_received(QByteArray bytes)
{
    //Debug::debug() << " ---- CoverProvider_Discogs -> slot_cover_received";

    // Get id from sender reply
    QObject* reply = qobject_cast<QObject*>(sender());
    if (!reply || !m_list_requests.contains(reply))  {
      emit search_finished(-1);
      return;
    }

    const int id = m_list_requests.take(reply);

    // Send signal for data reception
    emit data_received(id, bytes);
}

//! --  slot_cover_search_stop -------------------------------------------------
void CoverProvider_Discogs::slot_cover_search_stop()
{
    //Debug::debug() << " ---- CoverProvider_Discogs -> slot_cover_search_stop";

    // Get id from sender reply
    QObject* reply = qobject_cast<QObject*>(sender());
    if (!reply || !m_list_requests.contains(reply))  {
      emit search_finished(-1);
      return;
    }

    const int id = m_list_requests.take(reply);

    // Send signal for data reception
    emit search_finished(id);
}
